 /*
    Title:    
    Author:   
    Date:     
    Purpose:  
    needed
    Software: AVR-GCC
    needed
    Hardware: ATMEGA 16
    Note:     
*/
   

#include <avr\io.h>
#include <avr\interrupt.h>
#include <avr\signal.h>
#include <avr\eeprom.h>


typedef unsigned char  u08;
typedef signed char    s08;
typedef unsigned int   u16;
typedef struct {
		 u08 sec;
		 u08 min;
		 u08 hour;
		} time_rec;

#define TRUE 1
#define FALSE 0

///////////////////////////////////////////////////////////////////////////////////
// Software Versions-Nr.
///////////////////////////////////////////////////////////////////////////////////
//
#define version_hi	'1' 
#define version_lo	'1' 
#define version_rev	' ' 


///////////////////////////////////////////////////////////////////////////////////
// Die folgenden Werte werden zum einmaligen initialisieren der Uhr
// benutzt und knnen spter vom Benutzer per Software gendert werden
///////////////////////////////////////////////////////////////////////////////////
//
#define INITIAL_FALLBACK_TICKS  38	// Anzahl Ticks / 10 nach denen die Anzeige
					// auf die Tageszeit zurckfllt
				
#define INITIAL_DCF_SYNC_MINUTES 10	// Anzahl minuten nach denen die Zeit
					// neu synchronisiert wird
  
#define INITIAL_BUZZER_SECS	10	// Wecker-Dauer in sekunden

#define INITIAL_DARK_VALUE	3 

#define INITIAL_FADING_TIME	10
//
///////////////////////////////////////////////////////////////////////////////////


				
///////////////////////////////////////////////////////////////////////////////////
// Konstanten fr die DCF77 Decodierung
///////////////////////////////////////////////////////////////////////////////////
//
#define DCF_MIN_INIT_TICKS 192 // 1.5 sec
#define DCF_MAX_INIT_TICKS 255 // 2.0 sec
#define DCF_MAX_BIT_TICKS  28  // maximum 1/128 * 27 = 0.2109 sec
#define DCF_1BIT_TICKS  21     // minimum 1/128 * 22 = 0.1719 sec
#define DCF_0BIT_TICKS  9      // minimum 1/128 * 10 = 0.0781 sec 
#define DCF_IGNORE_TICKS  5    
// 
// 0er-Bits liegen blicherweise zwischen  11 und 13 Ticks 
// und 1er-Bits zwischen 23 und  25 Ticks
//
///////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////
// Konstanten fr die Betriebsstunden-Zhler
///////////////////////////////////////////////////////////////////////////////////
//
#define OPCOUNTER_TRIP		0
#define OPCOUNTER_TOTAL  	1

#define OPCOUNTER_MINUTES  	10
#define OPCOUNTER_SIZE		20
#define MAX_COUNTER_VALUE 	65532 
//
// In einer Integer Variablen werden 65532 Einheiten a 10 Minuten gezhlt
// Daraus ergibt sich:
//     65532 * 10 Minuten =  10922 Stunden
//     10922 Stunden * 20 = 218440 Stunden
//                        = ca. 9101 Tage
//                        = ca. 24,9 Jahre
// Es ergeben sich somit im Lauf von knapp 25 Jahren 65532
// Schreibzyklen fr den benutzten Speicherbereich
///////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////
// Die folgende Struktur bildet das benutzte EEPROM ab
///////////////////////////////////////////////////////////////////////////////////
//  
  
typedef struct
  { 
  	u08 ee_filler1;        // Fllfeld damit das erste Byte nicht beschrieben wird

  	u08 ee_valid;		// Enthlt diese Feld einen Wert != 1 wird das EEPROM neu
  				// initialisiert
  	
  	unsigned int op_counter[2][OPCOUNTER_SIZE]; // Betriebsstundenzhler 
  	u08 fallback_ticks;	// Zeit nach der die Anzeige auf Tageszeit zurckfllt
  	u08 sync_minutes;	// Minuten nach denen die Uhr neu synchronisiert wird
	
  	time_rec alarm_time;    // Weck-Zeit
  	u08 al_duration;	// Wecker Dauer in Sekunden

  	time_rec sw_on_time;    // Einschaltzeit
  	time_rec sw_off_time;   // Ausschaltzeit

        u08 dark_start1;   // Dunkelphase 1 Start
        u08 dark_end1;     // Dunkelphase 1 Ende
        
        u08 dark_start2;   // Dunkelphase 2 Start
        u08 dark_end2;     // Dunkelphase 2 Ende

        u08 dark_value1;         // Level fr Dunkelphase 1 (0-10); 
        u08 dark_value2;         // Level fr Dunkelphase 2 (0-10); 

        s08 dcf_offset;        // Offset in Stunden zur DCF-Zeit

        s08 fading_time;        // Dauer des Fadings
          
  } eeprom_data_rec;   // 

eeprom_data_rec eeprom_data;  

//
///////////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////////// 
// UART
///////////////////////////////////////////////////////////////////////////////////
//
#ifdef _AVR_IOM16_H_
#  define USR UCSRA
#  define UCR UCSRB
#  define UBRR UBRRL
#endif

#define UBRR_CONST	25	// Baudratenkonstante fr 9600 bei 4MHz Takt

#define UCR_CONST	0x98	// Enable Transmitter , Receiver , Receiver Interupt

#define UART_OBUF_SIZE 50	// Gre des Output-Buffers

u08 uart_obuf[UART_OBUF_SIZE];
u08 uart_obuf_first;
u08 uart_obuf_last;
u08 uart_obuf_len;
 
/*
#define UART_IBUF_SIZE 20    	// Gre des Input-Buffers

u08 uart_ibuf[UART_IBUF_SIZE];
u08 uart_ibuf_first = 0;
u08 uart_ibuf_last = 0;  
u08 uart_ibuf_len = 0;
*/

enum receive_modes   
	{rm_normal,  
	 rm_rreceived,
	 rm_r1received,
	 rm_wreceived,
	 rm_w1received,
	 rm_receivingdata
	 };

u08 receive_mode = rm_normal;

u08 pnum_buf[2];		// Buffer fr die Parameter-Nr bei R + W
u08 pnum;  			// Parameter-Nr

#define MAX_VALUE_LEN	8
u08 value_buf[MAX_VALUE_LEN];   
u08 value_len;
     
// Konstanten fr die Parameter-Nummern
#define	PNUM_COUNTER1		1	// RO Gesamtstundenzhler
#define	PNUM_COUNTER2		2	// R/W Stundenzhler (Schreiboperation fhrt zum Reset des Counters)
#define	PNUM_DAYTIME		3	// R/W Tageszeit
#define	PNUM_DATE		4	// R/W Datum
#define	PNUM_FALLBACK		5	// R/W Fallback-Ticks
#define	PNUM_SYNC_MIN		6	// R/W SyncMinutes
#define	PNUM_ALARM_TIME		7	// R/W Weckzeit
#define	PNUM_ALARM_SECS		8	// R/W Weck-Dauer
#define	PNUM_SW_ON		9	// R/W Schaltzeit EIN
#define	PNUM_SW_OFF		10	// R/W Schaltzeit AUS
#define	PNUM_DCF_SPECIAL	11	// RO Spezielle DCF77 Informationen
#define	PNUM_VERSION  		12	// RO Software-Version
#define	PNUM_DARK_START1        13	// R/W Beginn Dunkelphase 1
#define	PNUM_DARK_END1          14	// R/W Ende Dunkelphase 1
#define	PNUM_DARK_VALUE1        15	// R/W Wert fr Dunkelschaltung 1-5
#define	PNUM_DARK_START2        16	// R/W Beginn Dunkelphase 2 
#define	PNUM_DARK_END2          17	// R/W Ende Dunkelphase 2
#define	PNUM_DARK_VALUE2        18	// R/W Wert fr Dunkelschaltung 1-5
#define	PNUM_DCF_OFFSET		19	// R/W Offset zur DCF-Zeit
#define	PNUM_FADING_TIME	20	// R/W Offset zur DCF-Zeit
#define	MAX_PNUM	20
   
//
///////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////
// Fader / Blinker
///////////////////////////////////////////////////////////////////////////////////
//
#define  FADER_OLD	0
#define  FADER_NEW	1 

#define  BLINK_TICKER_COUNT 32

u08 fader_digits_old[6];
u08 fader_digits_new[6];
u08 fader_timer_main; 
u08 fader_timer_new;
u08 fader_timer_old;
u08 use_fader;        
// 
///////////////////////////////////////////////////////////////////////////////////
 
///////////////////////////////////////////////////////////////////////////////////
// Timer-Werte  
///////////////////////////////////////////////////////////////////////////////////
//
#define	T0PSC	   0x02 		// Timer0 Prescaler-Wert
#define	T0CV	   100  		// Timer0 Counter-Wert

u08 T1CV_TAB[11] = { 
	                          T0CV+((255-T0CV) * 1.0),
	                          T0CV+((255-T0CV) * 0.81),
	                          T0CV+((255-T0CV) * 0.64),
	                          T0CV+((255-T0CV) * 0.49),
	                          T0CV+((255-T0CV) * 0.36),
	                          T0CV+((255-T0CV) * 0.25),
	                          T0CV+((255-T0CV) * 0.16),
	                          T0CV+((255-T0CV) * 0.09),
	                          T0CV+((255-T0CV) * 0.05),
	                          T0CV+((255-T0CV) * 0.02),
	                          T0CV+((255-T0CV) * 0.0)
	                         };
 
u08 T1CV = T0CV;

#define	T2PSC	 0x01   	// Timer2 Prescaler-Wert
//
/////////////////////////////////////////////////////////////////////////////////// 

///////////////////////////////////////////////////////////////////////////////////
// Pin- und Port Mappings
///////////////////////////////////////////////////////////////////////////////////
// 
#define TUBE_SEL_PORT PORTD
#define TUBE_SEL_DIRP DDRD
#define TUBE_SEL_MASK 0xFC // 1111 1100
 
#define TUBE_VAL07_PORT PORTA
#define TUBE_VAL07_DIRP DDRA
#define TUBE_VAL07_MASK 0xFF // 1111 1111

#define TUBE_VAL89_PORT PORTC
#define TUBE_VAL89_DIRP DDRC
#define TUBE_VAL89_MASK 0x30 // 0011 0000 

#define LAMPS_PORT	PORTC 
#define LAMPS_DIRP	DDRC
#define	LAMP1_PIN	0x04  // 0000 0100
#define	LAMP2_PIN	0x08  // 0000 1000
#define NOT_LAMPS	0xF3  // 1111 0011 
#define	DCF_LAMP	LAMP2_PIN
#define	STAT_LAMP	LAMP1_PIN

#define BUTTONS_PORT	PORTB 
#define BUTTONS_PINREG 	PINB  
#define	BUTTON1_PIN	0x02 // 0000 0010	
#define	BUTTON2_PIN	0x04 // 0000 0100

#define DCF77_PORT	PORTB
#define DCF77_PINREG	PINB
#define DCF77_PIN	0x01

#define BUZZER_PORT	PORTB
#define BUZZER_DIRP	DDRB
#define BUZZER_PIN	4

#define SWITCH_PORT	PORTB
#define SWITCH_DIRP	DDRB
#define SWITCH_PIN	3
#define SWITCH_PIN_MASK	0x08 // 0000 1000

#define BLINK_NONE	TUBE_SEL_MASK
#define BLINK_56	0x3C	// 0011 1100
#define BLINK_34	0xCC	// 1100 1100
#define BLINK_12	0xF0	// 1111 0000
//
///////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////
// Variablen fr die Zustnde der Taster
///////////////////////////////////////////////////////////////////////////////////
//
#define BUTTON1	0
#define BUTTON2	1

#define BT_SHORT_TICKS	16	// Ticks die ein Button mindestenst 
	  		        // gedrckt sein mu
#define BT_LONG_TICKS	96	// Ticks die ein Button gedrckt sein 
				// mu um als lang gedrckt zu gelten
#define BT_REPEAT_TICKS	16	// Ticks die ein Button gedrckt sein 
				// mu um auto repeat auszulsen
enum button_states 
	{ bs_b12off,
	  bs_b1short, bs_b2short,
	  bs_b1long, bs_b2long, 
	  bs_b1repeatet, bs_b2repeatet,
	  bs_b12long, 
	  bs_blocked 
	  }; 

enum bt_timer_states 
	{ bt_b12off,bt_b1on,bt_b2on,bt_b12on };

u08 button_state;
u08 button_timer[2];
//
///////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////
// Variablen fr die verschiedenen Zustnde der Uhr
///////////////////////////////////////////////////////////////////////////////////
//
enum display_modes   
	{dm_time,
	 dm_date,
	 dm_totalophours,
	 dm_ophours,
	 dm_alarmtime,
	 dm_ontime,
	 dm_offtime};
	 
enum clock_modes
	{cm_run,
	 cm_set};
	 

enum dcf_states   
	{ ds_wait_for_gap,
	  ds_gap_received,
	  ds_minute_mark_received,
	  ds_receiving_bits
	  };
 
enum dcf_sync_states 
	{ ss_firstsync,
	  ss_resync,
	  ss_resync_after_error,
	  ss_insync,
	  ss_manual,
	  ss_no_sync_needed
	}; 

u08 display_mode;
u08 clock_mode;
 
u08 dcf_sync_state;
u08 dcf_receiving_state;
//
///////////////////////////////////////////////////////////////////////////////////


#define BT_TONE_TICKS	2

u08 ticker;
int fallback_ticker;
 


typedef struct {
    	         u08 hi;
  	         u08 lo;
  	       } bcd_bytes;
	
struct {
         bcd_bytes min;
         bcd_bytes hour;
         bcd_bytes day;
         bcd_bytes mon;
         bcd_bytes year;
       } dcf_data;

u08 dcf_completed;
u08 dcf_special_bits;
u08 dcf_first_bit_to_receive;
u08 dcf_second;
u08 dcf_bit;

u08 display_digits[6];
u08 display_mask; 
u08 blink_mask;
u08 blink_pair;

u08 lamp_stat;
 
time_rec the_time;
u08 day=1, mon=1, year;

u08 mdays[13] = {0,31,28,31,30,31,30,30,31,30,31,30,31};
 
u08 second_changed;
u08 ticker_changed;

u08 setting_seconds;

#define no_digit 0x0F

u08 buzzer_counter;
u08 button_tone_counter;   
 
u08 alarm_changed; 
u08 switch_time_changed;
u08 opcounter_min, opcounter_sec;

void invalidate_dcf_time(void);
void add_day(void);
void subtract_day(void);

// Makros statt Funktionen wg. Speicherverbrauch
#define button_tone_on  { cbi(BUZZER_PORT,BUZZER_PIN); button_tone_counter = 3; }
#define button_tone_off { sbi(BUZZER_PORT,BUZZER_PIN); button_tone_counter = 0;}
#define buzzer_on       { buzzer_counter = eeprom_data.al_duration; }
#define buzzer_off      { sbi(BUZZER_PORT,BUZZER_PIN); buzzer_counter = 0; }
#define switch_on       { sbi(SWITCH_PORT,SWITCH_PIN); }
#define switch_off      { cbi(SWITCH_PORT,SWITCH_PIN); }

#include "ndisplay.c"
#include "neeprom.c"
#include "nopcounter.c"
#include "nuart.c"


void lamp_on(u08 lamp)
{
  lamp_stat |= lamp;
}
 
void lamp_off(u08 lamp)
{
  lamp_stat &= ~lamp;
}

void lamp_invert(u08 lamp)
{
  lamp_stat ^= lamp;
}

/*
void clear_digits(void)

{
  display_digits[0] = no_digit;
  display_digits[1] = no_digit;
  display_digits[2] = no_digit;
  display_digits[3] = no_digit;
  display_digits[4] = no_digit;
  display_digits[5] = no_digit;
}  
 */

void decimal2digits(u08 decimal_value, u08 column, u08 zeroes)

{  
   u08 d1,d2;

   d1 = decimal_value / 10; 
   if (!d1 && !zeroes)
     d1 = no_digit;
   d2 = decimal_value % 10;
   if (use_fader)
   {
     set_fader_digit(column,d1);
     set_fader_digit(column+1,d2);
   }
   else
   {
     display_digits[column] = d1;
     display_digits[column+1] = d2;
   }
}
	

void setmode(u08 newmode)
 
{
  display_mode = newmode;
  fallback_ticker = eeprom_data.fallback_ticks * 10;
}

void ophours2display(void)

{ unsigned long ltemp;
  u08 show_zeroes;
  u08 d1,d2,d3;

  ltemp = get_opcounter_value(OPCOUNTER_TRIP);

  clear_fader(); 
  lamp_off(LAMP1_PIN);
  lamp_off(LAMP2_PIN);

  use_fader = FALSE;
  show_zeroes = FALSE;
  
  display_digits[0] = no_digit;
  display_digits[1] = no_digit;
  display_digits[2] = no_digit;
  display_digits[3] = no_digit;

  d3 = ltemp % 100;
  ltemp = ltemp / 100;
  d2 = ltemp % 100;
  ltemp = ltemp / 100;
  d1 = ltemp % 100;
 
  if (d1)
  {
    decimal2digits(d1,0,show_zeroes);
    show_zeroes = TRUE;
  }
  if (d2) 
  {
    decimal2digits(d2,2,show_zeroes);
    show_zeroes = TRUE;
  }
  decimal2digits(d3,4,show_zeroes);
  
  setmode(dm_ophours);
}


void timevalue2display(time_rec t1, u08 first_zero)

{
  clear_fader();

  decimal2digits(t1.hour,0,first_zero);
  decimal2digits(t1.min,2,1);
  decimal2digits(t1.sec,4,1);
}

void ontime2display(void)  

{
  use_fader = FALSE;
  timevalue2display(eeprom_data.sw_on_time,1);

  lamp_on(LAMP1_PIN);
  lamp_off(LAMP2_PIN); 

  setmode(dm_ontime);
}

void offtime2display(void)

{
  use_fader = FALSE;
  timevalue2display(eeprom_data.sw_off_time,1);

  lamp_off(LAMP1_PIN);
  lamp_on(LAMP2_PIN);

  setmode(dm_offtime); 
}

void alarmtime2display(void)
 
{
  use_fader = FALSE;
  timevalue2display(eeprom_data.alarm_time,1);

  lamp_on(LAMP1_PIN);
  lamp_on(LAMP2_PIN); 

  setmode(dm_alarmtime);
}


void time2display(void)  

{ 
  use_fader = ((clock_mode == cm_run) && (display_mode == dm_time));
  
  if ((dcf_sync_state == ss_insync) || (dcf_sync_state == ss_insync))
    lamp_on(STAT_LAMP);

  timevalue2display(the_time,0);

  display_mode = dm_time;
  
  if (use_fader)
    start_fader();
}


void date2display(void)

{ 
  clear_fader();
  use_fader = FALSE;
  decimal2digits(day,0,0);
  lamp_on(STAT_LAMP);
  decimal2digits(mon,2,0);
  lamp_on(DCF_LAMP);
  decimal2digits(year,4,1);

  setmode(dm_date);
}


u08 time_compare(time_rec t1, time_rec t2)

{ 
  if (t1.hour > t2.hour) return 1;
  if (t1.hour < t2.hour) return 2;
  
  if (t1.min > t2.min) return 1;
  if (t1.min < t2.min) return 2;
  
  if (t1.sec > t2.sec) return 1;
  if (t1.sec < t2.sec) return 2;
  
  return 0;
}

#define time_equal(t1,t2) time_compare(t1,t2) == 0
#define time_higher(t1,t2) time_compare(t1,t2) == 1
#define time_lower(t1,t2) time_compare(t1,t2) == 2
#define time_not_lower(t1,t2) time_compare(t1,t2) < 2

u08 time_between(time_rec t1, time_rec t2)

{  time_rec tmp_time;

  tmp_time = the_time;

  if (time_higher(t1,t2))
  {
    t2.hour += 24;
    if (time_higher(t1,tmp_time))
      tmp_time.hour += 24;
  }
  
  if ((time_not_lower(tmp_time,t1)) && (time_lower(tmp_time,t2)))
    return 1;
  return 0;
}
/*
u08 time_equal(time_rec t1, time_rec t2)

{ 
  if (t1.hour != t2.hour) return 0;
  if (t1.min != t2.min) return 0;
  if (t1.sec != t2.sec) return 0;
  
  return 1;
}

*/
u08 hour_between(u08 hour1, u08 hour2)

{ 
  u08 hourt;

  hourt = the_time.hour;
  if (hour1 > hour2) // Erste Zeit ist grer
  {
    hour2 += 24;
    if (hour1 >hourt)
      hourt += 24;
  }
  if ((hourt >= hour1) && (hourt < hour2))
    return 1;
  else 
    return 0;
}

void set_february_days(void)

{
  if (year % 4)
    mdays[2] = 28;
  else
    mdays[2] = 29;
}

void subtract_day(void)

{ 
  set_february_days();

  day--;
  if (!day)
  {
    mon--;
    if (!mon)
    {
      mon = 12;
      year--;
      if (year > 99)
        year = 99;
     	
    } 
    day = mdays[mon];
  }
}

void add_day(void)

{
  set_february_days();

  day++;
  if (day <= mdays[mon])
    return;
    
  day = 1;
    
  mon++;
  if (mon <= 12)
    return;
    
  mon = 1;
  year++;
  if (year > 99)
    year = 0;
}


void add_second(void)

{
  second_changed = 1;

  the_time.sec++;
  ticker = 0;
  if (the_time.sec < 60)
    return;
  
  the_time.sec = 0;
  the_time.min++;
  if (the_time.min < 60)
    return;
   
  the_time.min = 0;
  the_time.hour++;
  if (the_time.hour < 24)
    return;
    
  the_time.hour = 0;
  add_day();
}
  

void inc_timevalue(u08 tv_no, time_rec* t1)

{ 
  if (tv_no == 1)
  { 
    (*t1).hour++;
    if ((*t1).hour > 23)
      (*t1).hour = 0;
  }
 
  if (tv_no == 2)
  {
    (*t1).min++;
    if ((*t1).min > 59)
      (*t1).min = 0;
  }
 
  if (tv_no == 3)
  {
    (*t1).sec++;
    if ((*t1).sec > 59)
      (*t1).sec = 0;
  }
}


void inc_date(u08 dv_no)

{
  set_february_days();

  if (dv_no == 1)
  {
    day++;
    if (day > mdays[mon])
      day = 1;
  }
  
  if (dv_no == 2)
  { 
    mon++;
    if (mon > 12)
      mon = 1;
  }

  if (dv_no == 3)
  { 
    year++;
    if (year > 99)
      year = 0;
  }
}


void set_blinkpair(u08 pair)

{ blink_pair = pair;
  blink_mask = TUBE_SEL_MASK & ~(3 << (pair << 1));
} 


void init_setmode(void)

{
  alarm_changed = 0;	
  switch_time_changed = 0;	

  if (display_mode == dm_date)
    set_blinkpair(3);
  else
    set_blinkpair(1);
  clock_mode = cm_set;
}

u08 nextsetpair(void)

{
  if (display_mode == dm_date)
  {
    if (blink_pair > 1)
      set_blinkpair(blink_pair-1);
    else
      return 0;
    if (blink_pair == 1)
    {
      add_day();
      subtract_day();
      date2display();
    }
  }
  else
  {
    if (blink_pair < 3)
      set_blinkpair(blink_pair+1);
    else
      return 0;
  }

  return 1;
}

void button_actions_run(void)

{
  if (button_state != bs_b12off)
  {
    if (T1CV == 255) // wenn dunkel geschaltet ist
    {
      setmode(display_mode);  // setzt den fallback_ticker
      T1CV = T1CV_TAB[10];
      return;
    }
    
    if (buzzer_counter > 0)
    {
      buzzer_off;
      button_tone_on;
      return;
     }
  }

  if (button_state == bs_b12long)
  {
    if ((display_mode != dm_ophours) && (display_mode != dm_totalophours))
    {
      init_setmode();
      button_tone_on;
    }
    return;
  }

  if (button_state == bs_b1short)
  {
    switch (display_mode)
    {
      case dm_time:
        date2display();
        break;
      case dm_date:
        ophours2display();
        break;
      default:
          time2display();
    }
    button_tone_on;
  }
        

  if (button_state == bs_b2short)
  {
    switch (display_mode)
    {
      case dm_time:
        alarmtime2display();
        break;
      case dm_alarmtime:
        ontime2display();
        break;
      case dm_ontime:
        offtime2display();
        break;
      default:
          time2display();
    }
    button_tone_on;
  }
  
  // Halte den Anzeige-Modus solange Knopf gedrckt ist
  if ((button_state == bs_b1repeatet) || (button_state == bs_b2repeatet))
  {
    fallback_ticker = 128;
  }

}


void exit_setmode(void)

{
  set_blinkpair(0);
  if (setting_seconds)
    ticker = 0;
  setting_seconds = FALSE;
  clock_mode = cm_run;
  display_mode = dm_time; 
  time2display();
  
  if (alarm_changed)
  {
    eeprom_data_wt(&eeprom_data.alarm_time);
    alarm_changed = 0;
  }

  if (switch_time_changed)
  {
    eeprom_data_wt(&eeprom_data.sw_on_time);
    eeprom_data_wt(&eeprom_data.sw_off_time);
    switch_time_changed = 0;
  }
}


void button_actions_set(void)

{
  if (button_state == bs_b2short)
  {
    if (!nextsetpair())
    {
      if (display_mode == dm_ontime)
      {
        offtime2display();
        set_blinkpair(1);
      }
      else
        exit_setmode();
    }
    button_tone_on;
  }
  
  if ((button_state == bs_b1short) || (button_state == bs_b1long) || (button_state == bs_b1repeatet))
  {
    switch (display_mode)
    {
      case dm_time :
        inc_timevalue(blink_pair,&the_time);
        dcf_sync_state = ss_manual;
        dcf_receiving_state = ds_wait_for_gap;
        if (blink_pair == 3)
          setting_seconds = 1;
        time2display();
        break;
      case dm_date :
        inc_date(blink_pair);
        dcf_sync_state = ss_manual;
        dcf_receiving_state = ds_wait_for_gap;
        date2display();
        break;
      case dm_alarmtime :
        inc_timevalue(blink_pair,&eeprom_data.alarm_time);
        alarm_changed = 1;
        alarmtime2display();
        break;
      case dm_ontime :
        inc_timevalue(blink_pair,&eeprom_data.sw_on_time);
        switch_time_changed = TRUE;
        ontime2display();
        break;
      case dm_offtime :
        inc_timevalue(blink_pair,&eeprom_data.sw_off_time);
        switch_time_changed = TRUE;
        offtime2display();
        break;
    }
    button_tone_on;
  }
}

void get_buttonstate(void)

{ u08 temp;
  u08 last_state;
  u08 button_idx;
  
  temp = inp(BUTTONS_PINREG);
  temp = ~temp;
  temp &= (BUTTON1_PIN | BUTTON2_PIN);

  button_state = bs_b12off;				
  
  // Wenn die Taster gesperrt sind warte so
  // lange bis beide wieder losgelassen worden sind
  if ((button_state == bs_blocked) & (temp))
    return;


  // Stelle den letzten Zustand der Taster fest
  last_state =  bt_b12off;				
  if (button_timer[BUTTON1])					
  {							
    if (button_timer[BUTTON2])
      last_state = bt_b12on;
    else
      last_state = bt_b1on;
  } 
  else
    if (button_timer[BUTTON2])
      last_state = bt_b2on;
    
  
  button_idx = 0;

  // Kein Taster ist gedrckt
  if (!temp)
  {
    button_timer[BUTTON1] = 0;
    button_timer[BUTTON2] = 0;
    return;
  }
  
  if (temp == BUTTON2_PIN)
    button_idx = 1;
    
  // Nur ein Taster ist gedrckt
  if ((temp == BUTTON1_PIN) || (temp == BUTTON2_PIN))
  {
    // Kein Taster oder der andere Taster war vorher gedrckt
    if ((last_state == bt_b12off) || (last_state == (bt_b2on-button_idx)))
    {
      button_timer[BUTTON1+button_idx] = 1;
      button_timer[BUTTON2-button_idx] = 0;
    }
    
    // Der Taster war vorher auch gedrckt
    if (last_state == (bt_b1on+button_idx))		
    {
      button_timer[button_idx]++;
      
      // Der Taster wird kurz gedrckt 
      if (button_timer[button_idx] == BT_SHORT_TICKS)
        button_state = bs_b1short+button_idx;

      // Der Taster wird lange gedrckt 
      if (button_timer[button_idx] == BT_LONG_TICKS)
        button_state = bs_b1long+button_idx;

      // Der Taster wird festgehalten
      if (button_timer[button_idx] >= (BT_LONG_TICKS + BT_REPEAT_TICKS))
      {
        button_state = bs_b1repeatet+button_idx;
        button_timer[button_idx] = BT_LONG_TICKS;
      }
      
    }

    // Beide Taster waren vorher gedrckt
    // Blockiere die Taster-Funktionen bis beide wieder losgelassen sind
    if (last_state == bt_b12on)		
      button_state = bs_blocked;

    return;
  } 
  
  // Beide Taster sind gedrckt
  if (temp == (BUTTON1_PIN | BUTTON2_PIN))
  {

    button_timer[BUTTON1]++;
    button_timer[BUTTON2]++;

    // Taster 1 war gedrckt
    if (last_state == bt_b1on)
      button_timer[BUTTON2] = button_timer[BUTTON1];
    
    // Taster 2 war gedrckt
    if (last_state == bt_b2on)
        button_timer[BUTTON1] = button_timer[BUTTON2];
    
    // Beide waren vorher schon gedrckt
    if (last_state == bt_b12on)
      if (button_timer[BUTTON1] == BT_LONG_TICKS)
        button_state = bs_b12long;
  }
}


void test_buttons(void)

{  
  get_buttonstate();

  if (clock_mode == cm_run)
    button_actions_run();
  else
    button_actions_set();
    
}


void test_blinker(void)

{ static u08 blink_ticker;

  blink_ticker++;
  if (blink_ticker >= BLINK_TICKER_COUNT)
  {
    if (display_mask == TUBE_SEL_MASK)
      display_mask = blink_mask;
    else
      display_mask = TUBE_SEL_MASK;
    blink_ticker = 0;
  }
}


void test_buzzer(void)

{ static u08 buzz_ticker;

  if (button_tone_counter)
  { 
    button_tone_counter--;
    if (!button_tone_counter)
      button_tone_off;
  }

  if (!buzzer_counter)
    return;
  
  buzz_ticker++;
 
  if (buzz_ticker > 128)
    buzz_ticker = 1;
  
  if ((buzz_ticker == 8) || (buzz_ticker == 24))
    cbi(BUZZER_PORT,BUZZER_PIN);

  if ((buzz_ticker == 16) || (buzz_ticker == 32))
    sbi(BUZZER_PORT,BUZZER_PIN);
}


void test_fallback(void)
 
{
  if (fallback_ticker && (clock_mode != cm_set)) 
  {
    fallback_ticker--;
    if (!fallback_ticker)
    {
      blink_mask = BLINK_NONE;
      time2display();
    }
  } 
}

void add_dcf_offset(void)

{ s08 temp;

  temp = the_time.hour + eeprom_data.dcf_offset;

  if (temp < 0)
  {
    temp += 24;
    subtract_day();
  }
  if (temp > 23)
  {
    temp -= 24;
    add_day();
  }
  the_time.hour = temp;
}


void invalidate_dcf_time(void) 

{
  dcf_sync_state = ss_resync;
  dcf_receiving_state = ds_wait_for_gap;
  dcf_first_bit_to_receive = 1;
}


void dcf_receive_error(void) 

{
  if ((!dcf_first_bit_to_receive) && (dcf_sync_state != ss_firstsync))
    dcf_sync_state = ss_resync_after_error;
  dcf_receiving_state = ds_wait_for_gap;
  dcf_first_bit_to_receive = 0;
}


u08 bcd2byte(bcd_bytes bcd)

{
  return (bcd.hi * 10) + bcd.lo;
}

 
void process_dcf_bit(void)

{ static u08 crc;
  static u08 temp_special_bits;
  

  dcf_first_bit_to_receive = 0;
  
  if ((dcf_second != 28) && (dcf_second != 35) && (dcf_second != 58))
    crc += dcf_bit;
      
  if (dcf_second == 0)
  {
    if (dcf_completed)  // vorheriges telegramm komplett empfangen
    {
      the_time.min = bcd2byte(dcf_data.min);
      the_time.hour = bcd2byte(dcf_data.hour);
      the_time.sec = 0;   
      ticker = 13;
      day = bcd2byte(dcf_data.day);
      mon = bcd2byte(dcf_data.mon);
      year = bcd2byte(dcf_data.year);
      add_dcf_offset();
      dcf_special_bits = temp_special_bits;
      dcf_sync_state = ss_insync;
    }

     dcf_data.min.hi = 0;  dcf_data.min.lo = 0;
     dcf_data.hour.hi = 0; dcf_data.hour.lo = 0;
     dcf_data.day.hi = 0;  dcf_data.day.lo = 0;
     dcf_data.mon.hi = 0;  dcf_data.mon.lo = 0;
     dcf_data.year.hi = 0; dcf_data.year.lo = 0;
     
     temp_special_bits = 0;
     dcf_completed = 0;
     if (dcf_sync_state == ss_firstsync)
     {
       the_time.sec = 0;
       ticker = 13;
     }   
     if (display_mode == dm_time)
       time2display();
     if (display_mode == dm_date)
         date2display();
     return;
  }
  
  if ((dcf_second >= 15) && (dcf_second <= 19))  
  {
    temp_special_bits = temp_special_bits << 1;
    temp_special_bits += dcf_bit;
  }

  if (dcf_second == 20)            // Startbit fr Zeitinformation (mu 1 sein)
  {
    if (!dcf_bit)
      dcf_receiving_state = ds_wait_for_gap;
    crc = 0;
    return;
  }

  if ((dcf_second >= 21) && (dcf_second <= 24))		// Minute Lo
  {
    dcf_data.min.lo |= (dcf_bit << (dcf_second-21));
    return;
  }

  if ((dcf_second >= 25) && (dcf_second <= 27))		// Minute Hi
  {
    dcf_data.min.hi |= (dcf_bit << (dcf_second-25));
    return;
  }
   
  if (dcf_second == 28)  			// Erstes Prfbit
  {
    if ((crc & 0x01) != dcf_bit)
      dcf_receive_error();
    crc = 0;
    return;
  }

  if ((dcf_second >= 29) && (dcf_second <= 32))		// Stunde Lo
  {
    dcf_data.hour.lo |= (dcf_bit << (dcf_second-29));
    return;
  }

  if ((dcf_second >= 33) && (dcf_second <= 34))		// Stunde Hi
  {
    dcf_data.hour.hi |= (dcf_bit << (dcf_second-33));
    return;
  }

  if (dcf_second == 35)  			// Zweites Prfbit
  {						
  						
    if ((crc & 0x01) != dcf_bit)
    {
      dcf_receive_error();
      return;
    }

    if (dcf_sync_state == ss_firstsync) 	// Setzte schonmal die empfangene tageszeit
    {						// markiere aber noch nicht als gltig
      the_time.min = bcd2byte(dcf_data.min);
      the_time.hour = bcd2byte(dcf_data.hour);
      if (the_time.min == 0)
      {
        the_time.min = 59;
        if (the_time.hour == 0)
          the_time.hour = 23;
        else
          the_time.hour--;
      }
      else
        the_time.min--;
      the_time.sec = 35;
      add_dcf_offset();
      ticker = 13;
      if (display_mode == dm_time)
        time2display();
    }
    crc = 0;
    return;
  }

  if ((dcf_second >= 36) && (dcf_second <= 39))		// Tag Lo
  {
    dcf_data.day.lo |= (dcf_bit << (dcf_second-36));
    return;
  }

  if ((dcf_second >= 40) && (dcf_second <= 41))		// Tag Hi
  {
    dcf_data.day.hi |= (dcf_bit << (dcf_second-40));
    return;
  }


  if ((dcf_second >= 45) && (dcf_second <= 48))		// Monat Lo
  {
    dcf_data.mon.lo |= (dcf_bit << (dcf_second-45));
    return;
  }
       
  if (dcf_second == 49)		// Monat Hi
  {
    dcf_data.mon.hi = dcf_bit;
    return;
  }
  
  if ((dcf_second >= 50) && (dcf_second <= 53))		// Jahr Lo
  {
    dcf_data.year.lo |= (dcf_bit << (dcf_second-50));
    return;
  }

  if ((dcf_second >= 54) && (dcf_second <= 57))		// Jahr Hi
  {
    dcf_data.year.hi |= (dcf_bit << (dcf_second-54));
    return;
  }

  if (dcf_second == 58)  			// Drittes Prfbit
  {
    if ((crc & 0x01) != dcf_bit)
      dcf_receive_error();
    else
      dcf_completed = 1;
    return;
  }
}


void test_DCF(void)

{ static u08 dcf_ticks;
  static u08 dcf_last_state;
  u08 temp;

  if (!eeprom_data.sync_minutes)
    return;
    
  temp = inp(DCF77_PINREG);
  temp &= DCF77_PIN;
  if (display_mode == dm_time)
  {
    if ((temp) ||
       ((dcf_sync_state != ss_firstsync) & (dcf_sync_state != ss_resync_after_error))) 
      lamp_on(DCF_LAMP);
    else
      lamp_off(DCF_LAMP);
  }
  
  if (dcf_sync_state == ss_insync)  	// wenn synchron dann mach nix auer lampe setzten 
    return;

  if (dcf_ticks <= DCF_IGNORE_TICKS)  // ignoriere fr die ersten paar ticks nach pegelnderung
    temp = dcf_last_state;            // das signal um schwankungen rauszufiltern
    
  if (temp == dcf_last_state)	// keine pegel-nderung, ticks werden gezhlt 
  {			// nichts weiter zu tun
    dcf_ticks++;
    if (!dcf_ticks)    // berlauf, > 2 sek kein Empfang
      dcf_receive_error();
    return;
  }

  if (temp == 1) 	// wechseln von low nach high
  {
    if ((dcf_ticks >= DCF_MIN_INIT_TICKS)) // &&  (dcf_ticks <= DCF_MAX_INIT_TICKS))
      dcf_receiving_state = ds_gap_received;
  }
  else			// wechsel von high nach low
  {
    if ((dcf_ticks >= DCF_0BIT_TICKS) && (dcf_ticks <= DCF_MAX_BIT_TICKS))
    {
      if (dcf_ticks >= DCF_1BIT_TICKS)
        dcf_bit = 1;
      else 
        dcf_bit = 0;
    }
    else 
      dcf_receive_error();
        
    if (dcf_receiving_state == ds_gap_received) // berprfung auf minutenmarke 
    {
      if (dcf_bit == 0) // minutenmarke empfangen
        dcf_receiving_state = ds_minute_mark_received;
      else
        dcf_receive_error();
    } 
    
    if (dcf_receiving_state == ds_receiving_bits) // normale verarbeitung eines bits
    {
      dcf_second++;
      process_dcf_bit();
    }

    if (dcf_receiving_state == ds_minute_mark_received) // verarbeitung des ersten bits
    {
      dcf_second = 0;
      process_dcf_bit();
      dcf_receiving_state = ds_receiving_bits;
    }
  }

  dcf_last_state = temp;  
  dcf_ticks = 1; 
}



// Diese Funktion wird jede Sekunde aufgerufen
void second_loop(void)

{  u08 temp;
  
  second_changed = 0;
  
  // Die Glimmlampen blinken im Sekundentakt bei Anzeige der Weckzeit
  if (display_mode == dm_alarmtime)
  {
    lamp_invert(LAMP1_PIN);
    lamp_invert(LAMP2_PIN);
  } 

  // berprfe das Ende der Weck-Schaltung
  if (buzzer_counter > 0)
  {
    buzzer_counter--;
    if (buzzer_counter == 0)
      buzzer_off;
  }    
     
  // berprfe Weck- und Schalt- und Dunkel-Zeiten
  if ((clock_mode == cm_run) && (dcf_sync_state != ss_firstsync))

  {
  	
    if (eeprom_data.alarm_time.sec || eeprom_data.alarm_time.min || eeprom_data.alarm_time.hour) 
      if (time_equal(the_time,eeprom_data.alarm_time))
        buzzer_on;
/*
    if (time_equal(the_time,eeprom_data.sw_off_time)) 
      switch_off
    else
      if (time_equal(the_time,eeprom_data.sw_on_time)) 
        switch_on;
*/
    if (time_between(eeprom_data.sw_on_time,eeprom_data.sw_off_time)) 
      switch_on
    else
      switch_off;
    
    temp = 10;
  
    if (hour_between(eeprom_data.dark_start1,eeprom_data.dark_end1)) 
      temp = eeprom_data.dark_value1;
    if (hour_between(eeprom_data.dark_start2,eeprom_data.dark_end2))  
      temp = eeprom_data.dark_value2;
  
    if ((!temp) && (fallback_ticker || buzzer_counter)) // Wecken whrend Dunkelschaltung
      temp = 10;
    T1CV = T1CV_TAB[temp];
    
  } 
   
  // Zhle die Betriebsstunden in 10-Minuten Schritten
  if (T1CV < 255) // Rhren sind nicht aus
    opcounter_sec++;
  if (opcounter_sec == 60)
  {
    opcounter_min++; 
    opcounter_sec = 0;
  }
  if (opcounter_min == OPCOUNTER_MINUTES) 
  {
    increment_opcounters();
    opcounter_min = 0; 
  }

  // Re-Sync 1 min und 3 sec vor Sync-Zeit
  if ((the_time.sec == 57) && (dcf_sync_state == ss_insync))
  {
    if ((the_time.min + 2) % eeprom_data.sync_minutes == 0)
      invalidate_dcf_time();  
  }
 
  // Zeit zur Anzeige
  if (display_mode == dm_time)
  {
    time2display();
    if ((dcf_sync_state == ss_insync) || (dcf_sync_state == ss_resync))
      lamp_on(STAT_LAMP);  
    else
      lamp_invert(STAT_LAMP);
  }
    
  // Ausgabe der Zeit auf UART
  time2UART();
}

  
void ticker_loop(void)        
{ //u08 temp; 

  ticker_changed = 0;
  
  test_DCF();

  test_buttons();

  test_fallback();
  
  test_blinker();
  
  test_buzzer();
}


SIGNAL(SIG_OVERFLOW2)        
{ //u08 temp; 

  ticker++;
  ticker_changed = 1;  
  
  if (ticker > 127)
  { 
    ticker = 0;
    if (!setting_seconds)
      add_second();
  }
}


void set_pin_directions(void)

{
  u08 temp;
    
  temp = inp(TUBE_SEL_DIRP);  // pins fr die selektion als output
  temp |= TUBE_SEL_MASK;
  outp(temp,TUBE_SEL_DIRP);

//  temp = inp(TUBE_VAL07_DIRP);  // pins fr ziffern 0-7 als output
//  temp |= TUBE_VAL07_MASK;
  outp(TUBE_VAL07_MASK,TUBE_VAL07_DIRP);

  temp = inp(TUBE_VAL89_DIRP);  // pins fr ziffern 8-9 als output
  temp |= TUBE_VAL89_MASK;
  outp(temp,TUBE_VAL89_DIRP);

  temp = inp(LAMPS_DIRP);  // pins fr ziffern 8-9 als output
  temp |= LAMP1_PIN;	// pin fr lampe 1 als output
  temp |= LAMP2_PIN;	// pin fr lampe 2 als output
  outp(temp,LAMPS_DIRP);

  sbi(BUZZER_DIRP,BUZZER_PIN);	// pin fr den buzzer als output
  
  sbi(SWITCH_DIRP,SWITCH_PIN);	// pin fr den schalter als output

}   
    
   
void init_eeprom_data(void)
  
{ u08 i;
  char* j;    
 
  j = (char *)&eeprom_data;
  for (i=0;i<sizeof(eeprom_data);i++,j++)
    *j = 0;

  eeprom_data.fallback_ticks = INITIAL_FALLBACK_TICKS;
  eeprom_data.sync_minutes = INITIAL_DCF_SYNC_MINUTES;
  eeprom_data.al_duration = INITIAL_BUZZER_SECS;

  eeprom_data.dark_value1 = INITIAL_DARK_VALUE;
  eeprom_data.dark_value2 =INITIAL_DARK_VALUE;

  eeprom_data.fading_time = INITIAL_FADING_TIME;
  eeprom_data.ee_valid = 1;
}

void write_eeprom_data(void)
  
{ u08 i;
  char* j;    
 
  j = (char *)&eeprom_data;
  for (i=0;i<sizeof(eeprom_data);i++,j++)
    eeprom_data_wb(j); 
}

void read_eeprom_data(void)

{ u08 i, b1,b2,b3;
  char *j1;    

  j1 = (char*)&eeprom_data;
  for (i=0;i<sizeof(eeprom_data);i++,j1++)
  {  
    b1 = eeprom_rb(i+0);
    b2 = eeprom_rb(i+128);
    b3 = eeprom_rb(i+256);
    if (b1 != b2)
      if (b2 == b3)
    	b1 = b2;
    *j1 = b1;
  }
  
  if (eeprom_data.ee_valid != 1)  // Initialisierung des EEPROMS
    init_eeprom_data();
  
  // Schreibe eeprom bei jedem start
  write_eeprom_data();
}

int main( void )
{ u08 temp;         

  display_mask = BLINK_NONE;
  blink_mask = BLINK_NONE;  
  
  set_pin_directions();
  
  outp((1<<TOIE0) + (1<<TOIE1) + (1<<TOIE2), TIMSK); /* enable TCNT0-2 overflow */
  
  temp = inp(TCCR0);
  temp |= T0PSC;
  outp(temp,TCCR0);       // Prescaler Timer0
  outp(0,TCCR1B);          // Prescaler Timer1 (abgeschaltet)
  outp(T2PSC,TCCR2);       // Prescaler Timer2

  temp = inp(ASSR); 		// hole Async Status Register
  temp |= (1<<AS2);		// schalte AS2 ein
  outp(temp,ASSR);
  
   // Init UART
  outp(UBRR_CONST,UBRR);        // Baudraten-Konstante      
  outp(UCR_CONST,UCR);         // Transmitter + Receiver Enable 
  
  
  buzzer_off;		// mu initialisiert werden wg. invertierter Logik
  
  read_eeprom_data();

  if (!eeprom_data.sync_minutes) 
    dcf_sync_state = ss_no_sync_needed;
  else
    dcf_sync_state = ss_firstsync;

  sei();   
 
  for (;;)
  {
    if (second_changed)
      second_loop();
    if (ticker_changed)
      ticker_loop();
    check_uart_obuffer();
  }
}

